"""
Lucas, B. D., & Kanade, T. (1981). An iterative image registration technique with an application to stereo vision.


http://cseweb.ucsd.edu/classes/sp02/cse252/lucaskanade81.pdf
"""
import concurrent.futures
import logging
from enum import Enum
from typing import Optional
from typing import Tuple
from typing import Union

import cv2
import dask
import dask.array as da
import numpy as np
import tqdm
from dask.diagnostics import ProgressBar

from . import scaling
from . import utils

logger = logging.getLogger(__name__)


class Interpolation(str, Enum):
    none = "none"
    reshape = "reshape"
    rbf = "rbf"
    nearest = "nearest"


def default_options():
    return dict(
        winSize=(15, 15),
        maxLevel=2,
        interpolation=Interpolation.nearest,
        criteria=(cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT, 10, 0.03),
        step="auto",
    )


def resolve_step(step: Union[str, int], shape: Tuple[int, ...]) -> int:
    if isinstance(step, str):  # == "auto":
        step = max(int(min(shape) / 24), 1)

    return step


def flow(
    image: np.ndarray,
    reference_image: np.ndarray,
    points: Optional[np.ndarray] = None,
    winSize: Tuple[int, int] = (15, 15),
    maxLevel: int = 2,
    criteria=(cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT, 10, 0.03),
    step: Union[str, int] = "auto",
    interpolation: Interpolation = Interpolation.nearest,
) -> np.ndarray:
    """Compute the optical from reference_image to image

    Parameters
    ----------
    image : np.ndarray
        The target image
    reference_image : np.ndarray
        The reference image
    points : np.ndarray
        Points where to compute the motion vectors
    winSize : Tuple[int, int], optional
        Size of search window in each pyramid level, by default (15, 15)
    maxLevel : int, optional
        0-based maximal pyramid level number; if set to 0,
        pyramids are not used (single level), if set to 1,
        two levels are used, and so on; if pyramids are
        passed to input then algorithm will use as many
        levels as pyramids have but no more than maxLevel, by default 2
    criteria : tuple, optional
        Parameter, specifying the termination criteria of the iterative
        search algorithm (after the specified maximum number of iterations
        criteria.maxCount or when the search window moves by less than
        criteria.epsilon, by default (cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT, 10, 0.03)
    step : int, optional
        Step in pixels between points where motion is computed, by default 16
    interpolation : Interpolation
        Interpolate flow to original shape using radial basis function ('rbf'),
        nearest neigbour interpolation ('nearest') or do not interpolate but reshape ('reshape'),
        or use the original output from the LK algorithm ('none'), by default 'nearest'

    Returns
    -------
    np.ndarray
        Array of optical flow from reference image to image
    """
    if points is None:
        step = resolve_step(step, reference_image.shape)
        points = get_uniform_reference_points(reference_image, step=step)
    if image.dtype != np.uint8:
        image = utils.to_uint8(image)
    if reference_image.dtype != np.uint8:
        reference_image = utils.to_uint8(reference_image)

    f = _flow(image, reference_image, points, winSize, maxLevel, criteria)
    points = points.squeeze()

    if interpolation == Interpolation.none:
        return f

    if interpolation == Interpolation.rbf:
        return scaling.rbfinterp2d(
            points,
            f,
            np.arange(image.shape[1]),
            np.arange(image.shape[0]),
        )

    f = scaling.reshape_lk(points, f)
    if interpolation == Interpolation.nearest:
        new_shape: Tuple[int, int] = (
            reference_image.shape[0],
            reference_image.shape[1],
        )
        f = scaling.resize_vectors(f, new_shape)
    return f


def _flow(
    image: np.ndarray,
    reference_image: np.ndarray,
    points: np.ndarray,
    winSize: Tuple[int, int] = (15, 15),
    maxLevel: int = 2,
    criteria=(cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT, 10, 0.03),
) -> np.ndarray:
    """Compute the optical from reference_image to image

    Parameters
    ----------
    image : np.ndarray
        The target image
    reference_image : np.ndarray
        The reference image
    points : np.ndarray
        Points where to compute the motion vectors
    winSize : Tuple[int, int], optional
        Size of search window in each pyramid level, by default (15, 15)
    maxLevel : int, optional
        0-based maximal pyramid level number; if set to 0,
        pyramids are not used (single level), if set to 1,
        two levels are used, and so on; if pyramids are
        passed to input then algorithm will use as many
        levels as pyramids have but no more than maxLevel, by default 2
    criteria : tuple, optional
        Parameter, specifying the termination criteria of the iterative
        search algorithm (after the specified maximum number of iterations
        criteria.maxCount or when the search window moves by less than
        criteria.epsilon, by default (cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT, 10, 0.03)

    Returns
    -------
    np.ndarray
        Array of optical flow from reference image to image
    """

    if image.dtype != np.uint8:
        image = utils.to_uint8(image)
    if reference_image.dtype != np.uint8:
        reference_image = utils.to_uint8(reference_image)

    next_points, status, error = cv2.calcOpticalFlowPyrLK(
        reference_image,
        image,
        points,
        None,
        winSize=winSize,
        maxLevel=maxLevel,
        criteria=criteria,
    )

    flow = (next_points - points).reshape(-1, 2)

    return flow


def get_uniform_reference_points(image: np.ndarray, step: int = 48) -> np.ndarray:
    """Create a grid of uniformly spaced points width
    the gived steps size constraind by the image
    dimensions.

    Parameters
    ----------
    image : np.ndarray
        An image used to set the constraint on the dimensions
    step : int, optional
        The stepsize, by default 48

    Returns
    -------
    np.ndarray
        An array of uniformly spaced points
    """
    h, w = image.shape[:2]
    grid = np.mgrid[step // 2 : w : step, step // 2 : h : step].astype(int)
    return np.expand_dims(grid.astype(np.float32).reshape(2, -1).T, 1)


def get_displacements(
    frames,
    reference_image: np.ndarray,
    step: Union[str, int] = "auto",
    winSize: Tuple[int, int] = (15, 15),
    maxLevel: int = 2,
    criteria=(cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT, 10, 0.03),
    interpolation: Interpolation = Interpolation.nearest,
    **kwargs,
) -> np.ndarray:
    """Compute the optical flow using the Lucas Kanade method from
    the reference frame to all other frames

    Parameters
    ----------
    frames : np.ndarray
        The frames with some moving objects. Input must be of shape (N, M, T, 2), where
        N is the width, M is the height and  T is the number
    reference_image : np.ndarray
        The reference image
    step : int, optional
        Step in pixels between points where motion is computed, by default 16
    winSize : Tuple[int, int], optional
        Size of search window in each pyramid level, by default (15, 15)
    maxLevel : int, optional
        0-based maximal pyramid level number; if set to 0,
        pyramids are not used (single level), if set to 1,
        two levels are used, and so on; if pyramids are
        passed to input then algorithm will use as many
        levels as pyramids have but no more than maxLevel, by default 2
    criteria : tuple, optional
        Parameter, specifying the termination criteria of the iterative
        search algorithm (after the specified maximum number of iterations
        criteria.maxCount or when the search window moves by less than
        criteria.epsilon, by default (cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT, 10, 0.03)
    interpolation : Interpolation
        Interpolate flow to original shape using radial basis function ('rbf'),
        nearest neigbour interpolation ('nearest') or do not interpolate but reshape ('reshape'),
        or use the original output from the LK algorithm ('none'), by default 'nearest'

    Returns
    -------
    Array
        An array of motion vectors relative to the reference image. If shape of
        input frames are (N, M, T) then the shape of the output is (N', M', T', 2).
        Note if `resize=True` then we have (N, M, T, 2) = (N', M', T', 2).
    """
    if kwargs:
        logger.warning(f"Unknown arguments {kwargs!r} - ignoring")
    logger.info("Get displacements using Lucas Kanade")

    frames = utils.check_frame_dimensions(frames, reference_image)

    step = resolve_step(step, reference_image.shape)
    reference_points = get_uniform_reference_points(reference_image, step=step)

    num_frames = frames.shape[-1]

    all_flows = []
    for im in np.rollaxis(frames, 2):
        all_flows.append(
            dask.delayed(_flow)(
                im,
                reference_image,
                reference_points,
                winSize,
                maxLevel,
                criteria,
            ),
        )

    with ProgressBar():
        flows = da.stack(*da.compute(all_flows), axis=2)
    logger.info("Done with Lucas-Kanade method")

    if interpolation == Interpolation.none:
        return flows

    if interpolation == Interpolation.rbf:
        int_flows = np.zeros(
            (reference_image.shape[0], reference_image.shape[1], 2, num_frames),
        )
        p = reference_points.squeeze()
        x = np.arange(reference_image.shape[1])
        y = np.arange(reference_image.shape[0])
        int_args = ((p, f, x, y) for f in np.rollaxis(flows, 2))
        with concurrent.futures.ProcessPoolExecutor() as executor:
            for i, q in tqdm.tqdm(
                enumerate(executor.map(scaling.rbfinterp2d_map, int_args)),
                desc="Interpolate",
                total=num_frames,
            ):
                int_flows[:, :, :, i] = q
        return int_flows

    flows = scaling.reshape_lk(reference_points, flows)

    if interpolation == Interpolation.nearest:

        new_shape: Tuple[int, int] = (
            reference_image.shape[0],
            reference_image.shape[1],
        )

        flows = scaling.resize_vectors(flows, new_shape)
    return flows



import numpy as np
import pytest
from mps_motion import dualtvl1

docutils = pytest.importorskip("cv2.optflow")


def test_flow():

    reference_image = 255 * np.random.randint(0, 255, size=(64, 64), dtype=np.uint8)
    image = 255 * np.random.randint(0, 255, size=(64, 64), dtype=np.uint8)

    flow = dualtvl1.flow(image, reference_image)
    assert flow.shape == (reference_image.shape[0], reference_image.shape[1], 2)
    assert flow.dtype == np.float32



from unittest import mock

import dask.array as da
import numpy as np
import pytest
from mps_motion import FLOW_ALGORITHMS as _FLOW_ALGORITHMS
from mps_motion import Mechanics
from mps_motion import motion_tracking
from mps_motion import OpticalFlow
from mps_motion import utils

# from typing import Type
FLOW_ALGORITHMS = [alg for alg in dir(_FLOW_ALGORITHMS) if not alg.startswith("_")]


@pytest.mark.parametrize("reference_frame", ["min", "max", "median", "mean"])
def test_reference_frame_np(
    test_data: utils.MPSData,
    reference_frame: str,
):
    reference_image = motion_tracking.get_reference_image(
        reference_frame=reference_frame,
        frames=test_data.frames,
    )
    ref = getattr(np, reference_frame)(test_data.frames, axis=2)
    assert np.all(ref == reference_image)


@pytest.mark.parametrize(
    "reference_index, interval",
    [(0, (0, 3)), (1, (0, 3)), (2, (1, 4)), (-1, (-3, -1))],
)
def test_reference_frame_digit(
    test_data: utils.MPSData,
    reference_index: int,
    interval,
):

    reference_frame = test_data.time_stamps[int(reference_index)]
    if reference_index == -1:
        interval = (interval[0], test_data.time_stamps.size)

    ref = test_data.frames[:, :, interval[0] : interval[1]].mean(-1)
    reference_image = motion_tracking.get_reference_image(
        reference_frame=reference_frame,
        frames=test_data.frames,
        time_stamps=test_data.time_stamps,
    )
    assert np.all(abs(ref - reference_image) < 1e-12)


@pytest.mark.parametrize(
    "reference_index, interval",
    [("0", (0, 3)), ("1", (0, 3)), ("2", (1, 4)), ("-1", (-3, -1))],
)
def test_reference_frame_digit_str(
    test_data: utils.MPSData,
    reference_index: str,
    interval,
):
    reference_frame = test_data.time_stamps[int(reference_index)]

    if reference_index == "-1":
        interval = (interval[0], test_data.time_stamps.size)

    reference_image = motion_tracking.get_reference_image(
        reference_frame=reference_frame,
        frames=test_data.frames,
        time_stamps=test_data.time_stamps,
    )
    ref = test_data.frames[:, :, interval[0] : interval[1]].mean(-1)
    assert np.all(abs(ref - reference_image) < 1e-12)


@pytest.mark.parametrize("reference_frame", ["a", "std", ""])
def test_reference_frame_invalid(
    test_data: utils.MPSData,
    reference_frame: str,
):
    with pytest.raises(ValueError):
        motion_tracking.get_reference_image(
            reference_frame=reference_frame,
            frames=test_data.frames,
        )


@pytest.mark.parametrize("flow_algorithm", ["", "dslkgm"])
def test_invalid_algoritm(
    test_data: utils.MPSData,
    flow_algorithm: _FLOW_ALGORITHMS,
):
    with pytest.raises(ValueError):
        OpticalFlow(test_data, flow_algorithm=flow_algorithm)


@pytest.mark.parametrize("flow_algorithm", FLOW_ALGORITHMS)
def test_get_displacement_lazy(
    test_data: utils.MPSData,
    flow_algorithm: _FLOW_ALGORITHMS,
):
    np.random.seed(1)
    arr = np.random.random((4, 5, 3, 2))  # (width, heighth, num_time_points)
    with mock.patch(f"mps_motion.{flow_algorithm}.get_displacements") as _mock:
        _mock.return_value = arr
        m = OpticalFlow(test_data, flow_algorithm=flow_algorithm)
        U = m.get_displacements()
        assert (U[:, :, 0, :] == arr[:, :, 0, :]).all().compute()

    _mock.assert_called_once()


@pytest.mark.parametrize("flow_algorithm", FLOW_ALGORITHMS)
def test_get_displacement(test_data: utils.MPSData, flow_algorithm: _FLOW_ALGORITHMS):
    m = OpticalFlow(test_data, flow_algorithm=flow_algorithm)
    disp = m.get_displacements()
    assert disp.shape == (test_data.size_x, test_data.size_y, test_data.num_frames, 2)


def test_get_displacement_unit(test_data: utils.MPSData):

    m = OpticalFlow(test_data)
    disp_px = m.get_displacements(unit="pixels")
    disp_um = m.get_displacements(unit="um", recompute=True)

    assert da.isclose(
        disp_um.x.mean().max(),
        disp_px.x.mean().max() * test_data.info["um_per_pixel"],
    )

    assert da.isclose(
        disp_um.max().max(),
        disp_px.max().max() * test_data.info["um_per_pixel"],
    )
    assert da.isclose(
        disp_um.norm().max().max(),
        disp_px.norm().max().max() * test_data.info["um_per_pixel"],
    )


@pytest.mark.parametrize(
    "unit",
    ["um", "pixels"],
)
def test_get_displacement_scale(
    test_data: utils.MPSData,
    unit: str,
):
    """Test that there are now exceptions raised"""
    m = OpticalFlow(test_data)
    u_full = m.get_displacements(unit=unit)
    u = m.get_displacements(unit=unit, scale=0.5)
    assert np.isclose(u.mean().max().compute(), u_full.mean().max().compute())


@pytest.mark.parametrize(
    "unit",
    ["um", "pixels"],
)
def test_get_velocity_scale(
    test_data: utils.MPSData,
    unit: str,
):
    """Test that there are now exceptions raised"""
    m = OpticalFlow(test_data)
    v_full = m.get_velocities(unit=unit)
    v = m.get_velocities(unit=unit, scale=0.5)
    u_full = m.get_displacements(unit=unit, scale=0.5)
    v_full_from_u = Mechanics(u_full, t=test_data.time_stamps).velocity()

    u = m.get_displacements(unit=unit)
    v_from_u = Mechanics(u, t=test_data.time_stamps).velocity()

    # We cannot expected these to be all equal, but they should be
    # of the save order, so a 5-10% difference is OK
    assert np.isclose(v.mean().max().compute(), v_full.mean().max().compute(), rtol=0.1)

    assert np.isclose(
        v_from_u.mean().max().compute(),
        v_full_from_u.mean().max().compute(),
        rtol=0.1,
    )
    assert np.isclose(
        v_from_u.mean().max().compute(),
        v.mean().max().compute(),
        rtol=0.1,
    )


def test_OpticalFlow_options(test_data: utils.MPSData):

    step = 4
    m = OpticalFlow(test_data, flow_algorithm=_FLOW_ALGORITHMS.lucas_kanade, step=step)
    assert m.options["step"] == step


def test_estimate_reference_index(synthetic_trace):

    t = synthetic_trace.t[:-1]
    v = v = synthetic_trace.v

    index = motion_tracking.estimate_referece_image_from_velocity(
        t=t,
        v=v,
        rel_tol=0.001,
    )
    assert index == 87



import logging
import weakref
from pathlib import Path
from typing import Optional

import dask.array as da
import numpy as np

from . import filters
from . import utils

try:
    import h5py

    has_h5py = True
except ImportError:
    has_h5py = False

logger = logging.getLogger(__name__)


@utils.jit(nopython=True)
def sum_along_time(arr, t, i, j):
    x = 0
    for k in range(t):
        x += (1 / t) * np.abs(arr[i, j, k, :, :, :] - arr[i, j, k, 1, 1, 1]).sum()
    return x


@utils.jit(nopython=True, parallel=True)
def reduce_sliding_window(arr):
    n, m, t = arr.shape[:3]
    new_arr = np.zeros((n, m))

    for i in utils.prange(n):
        for j in range(m):
            new_arr[i, j] = sum_along_time(arr, t, i, j)

    return new_arr


def close_file(h5file):
    if h5file is not None:
        h5file.close()


class FrameSequence:
    """Object for holding a sequence of frames
    For example a component of a Tensor / Vector
    FrameSeqnecu

    """

    def __init__(
        self,
        array: utils.Array,
        dx: float = 1.0,
        scale: float = 1.0,
        fill_value: float = 0.0,
    ):
        """Constructor

        Parameters
        ----------
        array : np.ndarray or dask array
            Frame sequence of shape
            (width, height, num_time_steps)
        dx : float
            The Physical size of one
            pixel in the Frame, by default 1.0. Note this can
            also incorporate translation from pixel size to
            physical size.
        scale : float
            Another factor that should be reflexted and averaging
        fill_value: float
            Value to fill in for bad values masked arrays if used,
            by default 0.0.

        """
        assert isinstance(array, (da.core.Array, np.ndarray))
        self._ns = np if isinstance(array, np.ndarray) else da
        self._array = array
        self.dx = dx
        self.scale = scale
        self._h5file = None
        self._fill_value = fill_value
        self._finalizer = weakref.finalize(self, close_file, self._h5file)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()

    def cleanup(self):
        return self._finalizer()

    def __getitem__(self, *args, **kwargs):
        return self.array.__getitem__(*args, **kwargs)

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        if self.scale != other.scale:
            return False
        if self.dx != other.dx:
            return False
        return self._ns.isclose(self.array, other.array).all()

    def _check_other(self, other):
        if isinstance(other, FrameSequence):
            other = other.array

        if not isinstance(other, (da.core.Array, np.ndarray)):
            raise TypeError(
                f"Invald type {type(other)}, exepcted FrameSequence or array",
            )
        if self.shape != other.shape:
            raise ValueError(
                f"Incompatible shape, got {other.shape}, expected {self.shape}",
            )
        return other

    def __add__(self, other):
        other = self._check_other(other)
        return self.__class__(self.array + other, dx=self.dx, scale=self.scale)

    def __sub__(self, other):
        other = self._check_other(other)
        return self.__class__(self.array - other, dx=self.dx, scale=self.scale)

    def __mul__(self, other):
        if not np.isscalar(other):
            raise TypeError(f"Can only multiply with a scalar value, got {type(other)}")

        return self.__class__(other * self.array, dx=self.dx, scale=self.scale)

    def __rmul__(self, other):
        if not np.isscalar(other):
            raise TypeError(f"Can only multiply with a scalar value, got {type(other)}")

        return self.__class__(other * self.array, dx=self.dx, scale=self.scale)

    @property
    def fill_value(self) -> float:
        """Value used for as a replacement for bad values
        in masked arrays. This will only be used if you
        apply a mask to the array"""
        return self._fill_value

    @fill_value.setter
    def fill_value(self, value: float) -> None:
        self._fill_value = value

    def filter(
        self,
        filter_type: filters.Filters = filters.Filters.median,
        size: int = 3,
        sigma: float = 1.0,
    ) -> "FrameSequence":
        """Apply a median filter"""

        return FrameSequence(
            array=filters.apply_filter(
                self._array,
                size=size,
                sigma=sigma,
                filter_type=filter_type,
            ),
            dx=self.dx,
            scale=self.scale,
        )

    def save(self, path: utils.PathLike) -> None:
        path = Path(path)

        if path.is_file():
            path.unlink()

        suffixes = [".h5", ".npy"]
        msg = f"Expected suffix to be one of {suffixes}, got {path.suffix}"
        assert path.suffix in suffixes, msg
        if path.suffix == ".h5":
            if not has_h5py:
                raise IOError("Cannot save to HDF5 format. Please install h5py")

            with h5py.File(path, "w") as f:
                dataset = f.create_dataset("array", data=self.array_np)
                attr_manager = h5py.AttributeManager(dataset)
                attr_manager.create("scale", str(self.scale))
                attr_manager.create("dx", str(self.dx))
        else:
            np.save(
                path,
                {
                    "array": self.array_np,
                    "scale": self.scale,
                    "dx": self.dx,
                },  # type:ignore
            )

    @classmethod
    def from_file(cls, path, use_dask=True):

        path = Path(path)
        if not path.is_file():
            raise IOError(f"File {path} foes not exist")

        suffixes = [".h5", ".npy"]
        msg = f"Expected suffix to be one of {suffixes}, got {path.suffix}"
        assert path.suffix in suffixes, msg
        data = {}
        h5file = None
        if path.suffix == ".h5":
            h5file = h5py.File(path, "r")
            try:
                if "array" in h5file:
                    if use_dask:
                        data["array"] = da.from_array(h5file["array"])
                    else:
                        data["array"] = h5file["array"][...]
                    data.update(
                        dict(
                            zip(
                                h5file["array"].attrs.keys(),
                                map(float, h5file["array"].attrs.values()),
                            ),
                        ),
                    )
                    data["dx"] = float(h5file["array"].attrs.get("dx", 1))
                    data["scale"] = float(h5file["array"].attrs.get("scale", 1))
            except Exception:
                h5file.close()
        else:
            data.update(np.load(path, allow_pickle=True).item())

            if use_dask:
                data["array"] = da.from_array(data["array"])

        if "array" not in data:
            if h5file is not None:
                h5file.close()
            raise IOError(f"Unable to load data from file {path}")

        obj = cls(**data)
        obj._h5file = h5file
        return obj

    def local_averages(self, N: int, background_correction: bool = False):
        """Compute averages in local regions

        Parameters
        ----------
        N : int
            Number of regions along major axis
        background_correction : bool
            If true apply background correction algorithm
            to remove drift.

        Returns
        -------
        np.ndarray
            The local averages
        """
        try:
            from mps.analysis import local_averages
        except ImportError as ex:
            msg = "Please install the mps package for computing local averages"
            raise ImportError(msg) from ex

        return local_averages(
            self.array_np,
            np.arange(self.shape[2]),
            background_correction=background_correction,
            N=N,
        )

    def amplitude_mask(self, threshold: float = 30.0) -> np.ndarray:
        logger.info("Find mask for filtering based on amplitude")
        x = np.lib.stride_tricks.sliding_window_view(self.array_np, (3, 3, 3)).squeeze()
        y = reduce_sliding_window(x)
        mask = np.zeros(self.shape[:2], dtype=bool)
        mask[1:-1, 1:-1][y > threshold] = True
        return mask

    def apply_mask(self, mask: np.ndarray) -> None:
        logger.debug("Apply mask")
        self._array = da.ma.masked_array(
            self.array,
            np.tile(mask[..., np.newaxis], self.shape[2:]),
            fill_value=0.0,
        )

    def threshold(self, vmin: Optional[float] = None, vmax: Optional[float] = None):
        array = filters.threshold(self.array, vmin, vmax)
        return self.__class__(array, self.dx, self.scale)

    @property
    def array(self) -> utils.Array:
        return self._array

    @property
    def array_np(self) -> np.ndarray:
        array = self.array
        if isinstance(self._array, da.core.Array):
            array = self.array.compute()  # type: ignore
        return utils.unmask(array, fill_value=self.fill_value)

    @property
    def dx(self):
        return self._dx

    @dx.setter
    def dx(self, dx):
        assert dx > 0, "dx has to be positive"
        self._dx = dx

    @property
    def scale(self):
        return self._scale

    @scale.setter
    def scale(self, scale):
        assert scale > 0, "scale has to be positive"
        self._scale = scale

    @property
    def original_shape(self):
        w, h = self.shape[:2]
        return (int(w * self.dx), int(h * self.dx), *self.shape[2:])

    @property
    def shape(self):
        return self.array.shape

    def mean(self) -> utils.Array:
        return self.array.mean((0, 1)) * self.scale

    def max(self) -> utils.Array:
        return self.array.max(2) * self.scale

    def min(self) -> utils.Array:
        return self.array.min(2) * self.scale

    def compute(self) -> utils.Array:
        return self.array_np * self.scale

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.array.shape}, dx={self.dx}, scale={self.scale})"


class VectorFrameSequence(FrameSequence):
    """Object for holding a sequence of vectors.
    For example displacement

    """

    def __init__(self, array: utils.Array, dx: float = 1.0, scale: float = 1.0):
        """Constructor

        Parameters
        ----------
        array : np.ndarray or dask array
            Tensor sequence of shape
            (width, height, num_time_steps, 2)
        dx : float
            The Physical size of one
            pixel in the Frame, by default 1.0. Note this can
            also incorporate translation from pixel size to
            physical size.
        scale : float
            Another factor that should be reflexted and averaging

        """
        super().__init__(array, dx, scale)
        assert len(array.shape) == 4
        assert array.shape[3] == 2

    def filter(
        self,
        filter_type: filters.Filters = filters.Filters.median,
        size: int = 3,
        sigma: float = 1.0,
    ) -> "VectorFrameSequence":
        """Apply a filter"""

        array = filters.filter_vectors_par(
            self._array,
            size=size,
            sigma=sigma,
            filter_type=filter_type,
        )

        return VectorFrameSequence(array=array, dx=self.dx, scale=self.scale)

    def spline_smooth(self):
        arr = filters.spline_smooth(self.array)
        return VectorFrameSequence(da.from_array(arr), scale=self.scale, dx=self.dx)

    # def cartToPolar(self):

    #     mag, ang = cv2.cartToPolar(flow[...,0], flow[...,1], angleInDegrees=True)

    def apply_mask(self, mask: np.ndarray) -> None:
        self._array = da.ma.masked_array(
            self.array,
            np.tile(mask[..., np.newaxis, np.newaxis], self.shape[2:]),
            fill_value=0.0,
        )

    def threshold_norm(
        self,
        vmin: Optional[float] = None,
        vmax: Optional[float] = None,
    ) -> "VectorFrameSequence":
        array = filters.threshold_norm(self.array, self._ns, vmin, vmax)
        return VectorFrameSequence(array, self.dx, self.scale)

    def norm(self) -> FrameSequence:
        return FrameSequence(
            self._ns.linalg.norm(self._array, axis=3),
            dx=self.dx,
            scale=self.scale,
        )

    def angle(self) -> FrameSequence:
        return FrameSequence(
            self._ns.arctan2(self._array[:, :, :, 0], self._array[:, :, :, 1]),
            dx=self.dx,
            scale=self.scale,
        )

    @property
    def x(self) -> FrameSequence:
        return FrameSequence(self._array[:, :, :, 0], dx=self.dx, scale=self.scale)

    @property
    def y(self) -> FrameSequence:
        return FrameSequence(self._array[:, :, :, 1], dx=self.dx, scale=self.scale)


class TensorFrameSequence(FrameSequence):
    """Object for holding a sequence of tensors
    For example Green-Lagrange strain tensor
    and Cauchy stress tensor

    """

    def __init__(self, array: utils.Array, dx: float = 1.0, scale: float = 1.0):
        """Constructor

        Parameters
        ----------
        array : np.ndarray or dask array
            Tensor sequence of shape
            (width, height, num_time_steps, 2, 2)
        dx : float
            The Physical size of one
            pixel in the Frame, by default 1.0. Note this can
            also incorporate translation from pixel size to
            physical size.
        scale : float
            Another factor that should be reflexted and averaging

        """
        super().__init__(array, dx, scale)
        self._ns = np if isinstance(array, np.ndarray) else da
        assert len(array.shape) == 5
        assert array.shape[3] == array.shape[4] == 2

    def norm(self) -> FrameSequence:
        return FrameSequence(
            self._ns.linalg.norm(self._array, axis=(3, 4)),
            dx=self.dx,
            scale=self.scale,
        )

    @property
    def x(self) -> FrameSequence:
        return FrameSequence(self._array[:, :, :, 1, 1], dx=self.dx, scale=self.scale)

    @property
    def y(self) -> FrameSequence:
        return FrameSequence(self._array[:, :, :, 0, 0], dx=self.dx, scale=self.scale)

    @property
    def xy(self) -> FrameSequence:
        return FrameSequence(self._array[:, :, :, 1, 0], dx=self.dx, scale=self.scale)

    @property
    def yx(self) -> FrameSequence:
        return FrameSequence(self.array[:, :, :, 0, 1], dx=self.dx, scale=self.scale)

    def compute_eigenvalues(self) -> VectorFrameSequence:
        return VectorFrameSequence(
            np.linalg.eigvalsh(self.array_np),
            dx=self.dx,
            scale=self.scale,
        )
